import os

import numpy as np
from scipy.ndimage import median_filter, binary_fill_holes
from skimage import restoration, img_as_ubyte, filters, exposure
from skimage.filters import frangi, thresholding
from skimage.io import imread, imsave
from skimage.measure import regionprops, label
from skimage.morphology import disk, binary_erosion
from tqdm import tqdm


def subtract_background(img, radius=70):
    """
    Subtract background from image using rolling ball algorithm
    :param img: The image to subtract the background from
    :param radius: The radius of the rolling ball
    :return: The background subtracted image
    """
    normalized_radius = radius // 255
    kernel = restoration.ellipsoid_kernel(
        (radius * 2, radius * 2),
        normalized_radius * 2
    )
    rolling_ball = restoration.rolling_ball(img, radius=radius, kernel=kernel)
    return img - rolling_ball


def frangi_filter(img, sigma_range, black_ridges):
    """
    Apply the Frangi filter to an image
    :param img: The image to apply the filter to
    :param sigma_range: The range of sigmas to use (start, stop, step)
    :param black_ridges: Whether to detect black ridges
    :return: The filtered image
    """
    return frangi(img, sigmas=[x for x in range(*sigma_range)], black_ridges=black_ridges)


def li_threshold_image(img):
    """
    Apply the Li thresholding algorithm to an image
    :param img: The image to apply the thresholding to
    :return: The thresholded image
    """
    return img > thresholding.threshold_li(img, initial_guess=np.quantile(img, 0.95))


def sauvola_threshold_image(img, window_size=15):
    """
    Apply the Sauvola thresholding algorithm to an image
    :param img: The image to apply the thresholding to
    :param window_size: The size of the window to use for thresholding
    :return: The thresholded image
    """
    return img > filters.threshold_sauvola(img, window_size=window_size)


def estimate_tissue_mask(img):
    """
    Estimate the tissue mask from an image
    :param img: The image to estimate the mask from
    :return: The tissue mask
    Based on function found here: https://github.com/joe-from-mtl/sbhassisant-2d-3d-registration
    """
    mask_data = img > 0
    # Get a tissue threshold value
    threshold_tissue = thresholding.threshold_triangle(img[mask_data])

    # Apply threshold
    mask = img > threshold_tissue

    # Filter out noisy segmentation
    mask = median_filter(mask, 5)

    # Filter out small structures
    img_labels = label(mask)
    props = regionprops(img_labels)

    # Area threshold
    img_size = img.size
    tissue_labels = []
    for this_region in props:
        if this_region.area / img_size >= 0.05:
            tissue_labels.append(this_region.label)

    mask = np.zeros_like(mask)
    for this_label in tissue_labels:
        mask[img_labels == this_label] = 1

    return mask


def erode_mask(mask, disk_size=30):
    """
    Erode the outer edge of a mask
    :param mask: The mask to erode
    :param disk_size: The size of the disk to use for erosion
    :return: The eroded mask
    """
    return binary_erosion(mask, disk(disk_size))


def segment_2d_images(base_directory, images, erode_mask_size=30, background_filter_size=70,
                      frangi_sigma_range=(2, 16, 2), frangi_black_ridges=False, local_threshold=False,
                      local_threshold_size=15, erosion=True):
    """
    Segment 2D images. Finished files are not returned due to memory concerns.

    :param erosion: Enable or disable erosion of the mask
    :param base_directory: The base directory to save the results to and load the images from
    :param images: The filenames of the images to segment
    :param erode_mask_size: The size of the disk to use for erosion
    :param background_filter_size: The size of the background filter
    :param frangi_sigma_range: The range of sigmas to use for the Frangi filter (start, stop, step)
    :param frangi_black_ridges: Whether to detect black ridges
    :param local_threshold: Wheter to apply local or global thresholding
    :param local_threshold_size: The size of the window to use for local thresholding
    """
    if not os.path.exists(base_directory + "output/"):
        os.mkdir(base_directory + "output/")
    pbar = tqdm(images)
    for image in pbar:
        # Read image
        img = imread(base_directory + image, as_gray=True)

        pbar.set_description("Enhancing Contrast")
        # Enhance contrast
        img_enhanced = exposure.equalize_adapthist(img)

        pbar.set_description("Creating mask")
        # Create full mask
        mask = estimate_tissue_mask(img)

        pbar.set_description("Eroding mask")
        # Erode outer edge
        if erosion is True:
            mask_filled = binary_fill_holes(mask)
            erode = erode_mask(mask_filled, disk_size=erode_mask_size)
            mask = erode * mask
            del erode

        pbar.set_description("Removing background")
        # Remove background from image
        bg_less = subtract_background(img_enhanced, background_filter_size)

        pbar.set_description("Applying Frangi filter")
        # Apply Frangi filter
        frangi = frangi_filter(bg_less, frangi_sigma_range, frangi_black_ridges)

        pbar.set_description("Applying threshold")
        # Apply threshold
        if local_threshold:
            vessel_mask_raw = sauvola_threshold_image(frangi, local_threshold_size)
        else:
            vessel_mask_raw = li_threshold_image(frangi)

        pbar.set_description("Apply erosion to mask")
        # Apply erosion
        vessel_mask = vessel_mask_raw * mask

        pbar.set_description("Saving images")
        # Save image
        imsave(base_directory + "output/" + image + '_mask.tif', img_as_ubyte(mask), check_contrast=False)
        imsave(base_directory + "output/" + image + '_vessel_mask.tif', img_as_ubyte(vessel_mask), check_contrast=False)
        imsave(base_directory + "output/" + image + '_frangi.tif', frangi, check_contrast=False)
        # Clean memory
        del img, mask, bg_less, frangi, vessel_mask_raw, vessel_mask



import os

import h5py as h5
import nibabel as nib
import nrrd
import numpy as np
import tqdm as tqdm
import zarr
from ome_zarr.io import parse_url
from ome_zarr.writer import write_image


def get_first_key(file):
    """
    Get the first key in a HDF5 file.
    :param file: The HDF5 file.
    :return: The first key. Useful for getting the shape of the data.
    """
    keys = file.keys()
    for first_key in keys:
        first_key = first_key
        break
    return first_key


def load_hdf5(hdf5_file, map_file="temp.dat"):
    """
    Load a HDF5 into a numpy memmap.
    :param hdf5_file: The HDF5 file to load.
    :param map_file: The memmap file to save to.
    :return: The memmap.
    """
    with h5.File(hdf5_file) as file:
        keys = file.keys()
        n_frames = len(keys)
        key = get_first_key(file)
        frame = file[key][:]
        data = np.memmap(map_file, dtype=np.uint16, mode='w+', shape=(n_frames, frame.shape[0], frame.shape[1]))
        for i, key in enumerate(tqdm.tqdm(keys, desc="Converting HDF5 to zarr file")):
            frame = file[key][:]
            data[i, :, :] = frame
    return data


def convert_hdf5_to_nifti(hdf5_file, nifti_file):
    """
    Convert a HDF5 file from the lightsheet microscope to a NIFTI file.
    :param hdf5_file: Path to the HDF5 file.
    :param nifti_file: Path to the NIFTI file.
    :return:
    """
    map_file = "temp.dat"
    data = load_hdf5(hdf5_file, map_file)

    print("Transposing...")
    data = np.transpose(data, (1, 2, 0))
    print("Saving...")
    ni_img = nib.Nifti1Image(data, affine=np.eye(4), dtype=np.uint16)
    nib.save(ni_img, nifti_file)
    print("Done!")
    os.remove(map_file)


def save_zarr(data, zarr_file, axes="zxy", chunks=(1, 2048, 2048)):
    """
    Save a numpy array to a zarr file.
    :param data: The data to save.
    :param zarr_file: The zarr file to save to.
    :param axes: The order of the axes of the data.
    :param chunks: The chunk size to use.
    :return:
    """
    print("Saving...")
    os.mkdir(zarr_file)
    store = parse_url(zarr_file, mode="w").store
    root = zarr.group(store=store)
    write_image(image=data, group=root, axes=axes, storage_options=dict(chunks=chunks))
    print("Done!")


def convert_hdf5_to_zarr(hdf5_file, zarr_file):
    """
    Convert a HDF5 file from the lightsheet microscope to a zarr file.
    :param hdf5_file: Path to the HDF5 file.
    :param zarr_file: Path to the zarr file.
    :return:
    """
    map_file = "temp.dat"
    data = load_hdf5(hdf5_file, map_file)

    save_zarr(data, zarr_file, chunks=(1, 2048, 2048))
    os.remove(map_file)


def convert_nifti_to_zarr(nifti_file, zarr_file, shape=None):
    """
    Convert a NIFTI file to a zarr file.
    :param nifti_file: The NIFTI file to convert.
    :param zarr_file: The zarr file to save to.
    :param shape: The shape of the data. If None, the shape will be inferred from the NIFTI file.
    """
    print("Loading...")
    ni_img = nib.load(nifti_file)
    data = ni_img.get_fdata()
    if shape is None:
        shape = data.shape
    save_zarr(data, zarr_file, axes="xyz", chunks=shape)


def convert_nrrd_to_zarr(nrrd_file, zarr_file, shape=None):
    """
    Convert a NRRD file to a zarr file.
    :param nrrd_file: The NRRD file to convert.
    :param zarr_file: The zarr file to save.
    :param shape: The shape of the data to save. If None, the shape of the data will be used.
    """
    print("Loading...")
    data, header = nrrd.read(nrrd_file)
    if shape is None:
        shape = data.shape
    save_zarr(data, zarr_file, axes="xyz", chunks=shape)
